# pylint: disable=line-too-long,import-error,too-few-public-methods
"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import re
import json
import boto3


class GoogleKnowledgeGraphWriter:
    """GoogleKnowledgeGraphReader Class"""

    def __init__(self, bucket, folder_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket: str = bucket
        self.folder_path: str = folder_path
        self.success = False

    def is_success(self) -> None:
        """Update Instance Success Status to True"""
        self.success = True

    def write_to_s3(self, payload):
        """
        This pulls the yielded dataset from the Google Knowledge Graph reader
        and writes it to s3 so that duplication does not occur.
        :param payload: This is a key value object that looks like:
                        {
                            "data": list[dict, dict],
                            "date": string
                        }
        """

        # confirm the payload keys are matching accurately with what is expected
        if not {"data", "date"}.issubset(set(payload.keys())):
            raise KeyError("Invalid payload ensure you have date and data as keys")

        if not re.search(r"\d{4}\-\d{2}\-\d{2}", payload["date"]):
            raise ValueError(
                "Invalid date format for partitioning expected: 'YYYY-mm-dd'"
            )

        if not isinstance(payload["data"], list):
            raise TypeError("Invalid data passed: expected List[Dict, Dict]")

        _date = payload["date"].replace("-", "")
        _data = payload["data"]
        write_path = f"{self.folder_path}/{_date}.json"
        if _data:
            print(
                f"Writing data to s3://{self.bucket}/{write_path} partitioned by date."
            )
            self.s3_resource.Object(self.bucket, write_path).put(Body=json.dumps(_data))
            # this helps with unittesting so we set success to True
            self.is_success()



# pylint: disable=too-many-nested-blocks,too-many-locals,no-name-in-module,import-error,broad-except,lost-exception,inconsistent-return-statements,too-many-branches,too-many-statements
"""
ZOHO CRM READER MODULE
"""
import os
import time

from zcrmsdk import ZCRMRestClient, ZCRMException
from zcrmsdk.OAuthClient import ZohoOAuth
from zcrmsdk.Operations import ZCRMModule

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler
from sdc_dp_helpers.zoho.writers import CustomS3JsonWriter


class CustomZohoReader:
    """
    Custom Zoho Reader
    """

    def __init__(self, **kwargs):
        self.config = load_file(kwargs.get("config_path", None), "yaml")
        self.creds = load_file(kwargs.get("creds_path", None), "yaml")

        self.zoho_token = self.initialize_zoho_client()
        self.partial_pull = self.config.get("partial_pull", True)
        self.partial_pull_size = self.config.get("partial_pull_size", 5)

        self.module = self.config.get("module")

        self.writer = CustomS3JsonWriter(bucket=kwargs.get("bucket"))

    def initialize_zoho_client(self):
        """
        Zoho Client will be creating local auth and log files every time it is run.
        This method handles the directories and environments as well as the OAth Client.
        """
        # create file for app log without needing root and OS dependency
        current_dir = os.getcwd()
        try:
            app_log_file_path = self.creds.get("application_log_file_path").replace(
                "./", ""
            )
            os.mkdir(f"{current_dir}/{app_log_file_path}")
            log_file_path = f"{current_dir}/{app_log_file_path}/{'client_library.log'}"
            if not os.path.exists(log_file_path):
                with open(log_file_path, "a") as _file:
                    _file.write("")
        except FileExistsError:
            pass

        config = {
            "sandbox": self.creds.get("sandbox"),
            "applicationLogFilePath": self.creds.get(
                "application_log_file_path"
            ),  # a local log file is needed
            "client_id": self.creds.get("client_id"),
            "client_secret": self.creds.get("client_secret"),
            "redirect_uri": self.creds.get(
                "redirect_uri"
            ),  # this can be totally random
            "accounts_url": self.creds.get("accounts_url"),
            "token_persistence_path": self.creds.get("token_persistence_path"),
            "currentUserEmail": self.creds.get("current_user_email"),
        }

        ZCRMRestClient.initialize(config)
        return ZohoOAuth.get_client_instance()

    @staticmethod
    def match_keys(field_name, record):
        """
        Find keys in record and map them.
        """
        for key in record.keys():
            if field_name.lower() == key.lower():
                return key
        return False

    @staticmethod
    @request_handler(wait=0.2, backoff_factor=0.5, backoff_method="random")
    def record_response_handler(row, record):
        """
        Handle response data and add missing data to suit
        business requirements.
        """
        record["entity_id"] = row.entity_id

        if row.created_by is not None:
            record["created_by_id"] = row.created_by.id
        else:
            record["created_by_id"] = "null"

        if row.modified_by is not None:
            record["modified_by_id"] = row.modified_by.id
        else:
            record["modified_by_id"] = "null"

        if row.modified_by is not None:
            record["modified_by"] = row.modified_by.name
        else:
            record["modified_by"] = "null"

        if row.modified_time is not None:
            record["modified_time"] = row.modified_time
        else:
            record["modified_time"] = "null"

        if row.created_by is not None:
            record["created_by"] = row.created_by.name
        else:
            record["created_by"] = "null"

        if row.layout is not None:
            record["layout"] = row.layout.name
        else:
            record["layout"] = "null"

        if row.layout is not None:
            record["layout_id"] = row.layout.id
        else:
            record["layout_id"] = "null"

        record["owner_id"] = row.owner.id
        record["owner"] = row.owner.name
        record["created_time"] = row.created_time

        return record

    def record_run_query(self):
        """
        Run Zoho Client Query for Records based on module.
        """
        oauth_client = self.initialize_zoho_client()
        oauth_client.generate_access_token_from_refresh_token(
            self.creds.get("refresh_token"), self.creds.get("current_user_email")
        )

        module_instance = ZCRMModule.get_instance(self.module)
        all_fields = module_instance.get_all_fields().data

        page, per_page = 1, 200
        module_response = module_instance.get_records(
            None, "Modified_Time", "desc", page, per_page
        )
        try:
            if self.partial_pull:
                while page <= self.partial_pull_size:  # this is a partial pull
                    loop_count = 1
                    data_set = []
                    for row in module_response.data:
                        print(f"At loop: {loop_count}, and page: {page}.")
                        record = {}
                        for field in all_fields:
                            name = field.api_name
                            value = row.get_field_value(field.api_name)
                            record[name.lower()] = value
                        data_set.append(self.record_response_handler(row, record))
                        loop_count += 1

                    # write each page as a partition and clean data_set
                    print("Writing to S3.")
                    self.writer.write_to_s3(
                        json_data=data_set,
                        config=self.config,
                        partition=page,
                        layout="records",
                    )
                    page += 1

                print("No more data to import.")

            else:
                loop_count = 1
                data_set = []
                for row in module_response.data:
                    print(f"At loop: {loop_count}, and page: {page}.")
                    record = {}
                    for field in all_fields:
                        name = field.api_name
                        value = row.get_field_value(field.api_name)
                        record[name.lower()] = value
                    data_set.append(self.record_response_handler(row, record))
                    loop_count += 1

                # write entire set as single partition
                print("Writing to S3.")
                self.writer.write_to_s3(
                    json_data=data_set,
                    config=self.config,
                    partition=0,
                    layout="records",
                )
                page += 1

        except ZCRMException as ex:
            if ex.status_code == 204:
                print("No more data to import.")

        finally:
            return None

    def layouts_response_handler(self, module_object, layout, record):
        if self.module == "Products":
            record["layout_id"] = layout.id
            record[self.module + "_id"] = module_object.entity_id
            record["owner_id"] = module_object.owner.id
            record["owner_name"] = module_object.owner.name
            record["created_time"] = module_object.created_time
            record["created_by"] = module_object.created_by.name
            record["created_by_id"] = module_object.created_by.id
        elif self.module != "Products":
            record["layout_id"] = layout.id
            record[self.module + "_id"] = module_object.entity_id
            record["owner_id"] = module_object.owner.id
            record["owner_name"] = module_object.owner.name
            record["created_time"] = module_object.created_time
            record["created_by"] = module_object.created_by.name
            record["created_by_id"] = module_object.created_by.id
            record["modified_time"] = module_object.modified_time
            record["modified_by"] = module_object.modified_by.name
            record["modified_by_id"] = module_object.modified_by.id

            try:
                record["last_activity_time"] = module_object.last_activity_time
            except Exception:
                print("The last_activity_time is not available.")

            if self.module == "Deals":
                res = []
                for stage_record in module_object.get_relatedlist_records(
                    "Stage_History"
                ).data:
                    res.append(stage_record.field_data)
                record["Stage_History"] = res

        return record

    def layouts_run_query(self):
        oauth_client = self.initialize_zoho_client()
        oauth_client.generate_access_token_from_refresh_token(
            self.creds.get("refresh_token"), self.creds.get("current_user_email")
        )

        module_instance = ZCRMModule.get_instance(self.module)
        all_layouts = module_instance.get_all_layouts()

        for layout in all_layouts.data:
            if layout.id:
                # re-declare field_list for use in insert statement
                field_list = [
                    "layout_id",
                    self.module + "_ID",
                    "Owner_ID",
                    "Owner_Name",
                ]
                problem_fields = ["Tag", "Description"]
                if self.module == "Deals":
                    field_list.append("Stage_History")
                elif self.module == "Activities":
                    problem_fields = ["Tags"]

                # fetch fields by section inside the current layout
                for section in layout.sections:
                    if section is not None:
                        for field in (
                            section.fields if section.fields is not None else []
                        ):
                            if field.api_name not in problem_fields:
                                field_list.append(field.api_name)

                page, per_page = 1, 200
                module_response = module_instance.get_records(
                    None, "Modified_Time", "desc", page, per_page
                )
                try:
                    if self.partial_pull:
                        data_set = []
                        while page <= self.partial_pull_size:  # this is a partial pull
                            loop_count = 1
                            print(f"At loop: {loop_count}, and page: {page}.")
                            for module_object in module_response.data:
                                record = module_object.field_data

                                record = self.layouts_response_handler(
                                    module_object, layout, record
                                )

                                for field_name in field_list:
                                    if field_name not in record.keys():
                                        try:
                                            field_value = module_object.field_name
                                        except Exception:
                                            field_value = module_object.get_field_value(
                                                field_name
                                            )
                                        key_found = self.match_keys(
                                            field_name.lower(), record
                                        )
                                        if not key_found:
                                            record[field_name] = field_value

                                data_set.append(record)
                                loop_count += 1

                            # write each page as a partition and clean data_set
                            print("Writing to s3.")
                            self.writer.write_to_s3(
                                json_data=data_set,
                                config=self.config,
                                partition=page,
                                layout="layouts",
                            )
                            data_set = []

                            page += 1
                            # get_records dies very easily if overused,
                            # so permanent sleep method is recommended
                            time.sleep(1)
                            print("Waiting for 1 second.")

                        print("No more data to import.")

                    else:
                        # we can only run while no errors occur, else the API shuts down.
                        data_set = []
                        while 200 <= module_response.status_code < 300:
                            loop_count = 1
                            print(f"At loop: {loop_count}, and page: {page}.")
                            for module_object in module_response.data:
                                record = module_object.field_data

                                record = self.layouts_response_handler(
                                    module_object, layout, record
                                )

                                for field_name in field_list:
                                    if field_name not in record.keys():
                                        try:
                                            field_value = module_object.field_name
                                        except Exception:
                                            field_value = module_object.get_field_value(
                                                field_name
                                            )
                                        key_found = self.match_keys(
                                            field_name.lower(), record
                                        )
                                        if not key_found:
                                            record[field_name] = field_value

                                data_set.append(record)
                                loop_count += 1
                            page += 1

                            # get_records dies very easily if overused,
                            # so permanent sleep method is recommended
                            time.sleep(20)
                            print("Waiting for 20 seconds.")

                        print("Writing to s3.")
                        self.writer.write_to_s3(
                            json_data=data_set,
                            config=self.config,
                            partition=0,
                            layout="layouts",
                        )

                        print("No more data to import.")

                except ZCRMException as ex:
                    if ex.status_code == 204:
                        print("No more data to import.")

                finally:
                    return None

            else:
                print("No list id, skipping.")



"""
    CUSTOM READERS MODULE FOR GOOGLE SEARCH CONSOLE
"""
# pylint: disable=no-member,too-many-locals,broad-except,too-few-public-methods,arguments-differ
from typing import Any, List, Dict, Union
from googleapiclient import discovery
from google.oauth2 import service_account

from sdc_dp_helpers.base_readers import BaseReader

from sdc_dp_helpers.api_utilities.date_managers import date_range_iterator
from sdc_dp_helpers.api_utilities.file_managers import load_file


class GoogleSearchConsoleReader(BaseReader):
    """
    Google Search Console Reader v1
    """

    def __init__(self, creds_filepath: str, config_filepath: str):
        self.secrets: Dict[Any, Any] = load_file(creds_filepath)
        self.config: Dict[Any, Any] = load_file(config_filepath)
        self.service = self._get_auth()
        self.success: List[bool] = []

    def _get_auth(self):
        """
        Get our credentials initialised above and use those to get client

        """
        credentials = service_account.Credentials.from_service_account_info(
            info=self.secrets,
            scopes=["https://www.googleapis.com/auth/webmasters.readonly"],
        )
        service = discovery.build(
            serviceName="searchconsole", version="v1", credentials=credentials
        )
        return service

    def _normalize(
        self, response: Dict[Any, Any], query_dim_list: List[str]
    ) -> List[Dict[str, Any]]:
        """Method to normalize data into a flat structure and add metadata

        Args:
            response (Dict[Any, Any]): api response in dictionary format
            query_dim_list (List[str]): Zero or more dimensions to group results.

        Returns:
            Dict[str, Any]: flattened data dictionary
        """

        # added additional data that the api does not provide
        dataset_list: List[Dict[str, Any]] = []
        for _, row in enumerate(response["rows"]):
            row_dataset: Dict[str, Any] = {}
            row_dataset = {
                "site_url": self.config.get("site_url"),
                # keeping original search_type name for reporting
                "search_type": self.config.get("search_type"),
            }

            # get dimension data keys and values
            row_dataset.update(
                dict(
                    zip(
                        query_dim_list,
                        row.get("keys", []),
                    )
                )
            )

            # get metrics data
            for metric in self.config.get("metrics", []):
                row_dataset[metric] = row.get(metric)

            dataset_list.append(row_dataset)

        return dataset_list

    def _query_handler(
        self, start_date: str, end_date: str, start_row: int, query_dim_list: List[str]
    ) -> Union[Dict[Any, Any], None]:
        """Function to make api call

        Args:
            start_date (str): start of range inclusive 'YYYY-MM-DD'
            end_date (str): end of range inclusive 'YYYY-mm-dd'
            query_dim_list (List[str]): Zero or more dimensions to group results.

        Returns:
            Dict[Any, Any]: dictionary of response
        """
        response: Union[Dict[Any, Any], None] = None
        try:
            request = {
                "startDate": start_date,
                "endDate": end_date,
                "dimensions": query_dim_list,
                "metrics": self.config.get("metrics"),
                "searchType": self.config.get("search_type"),
                "rowLimit": self.config.get("row_limit", 25000),
                "startRow": start_row * self.config.get("row_limit", 25000),
                "aggregationType": self.config.get("aggregation_type", "auto"),
                "dimensionFilterGroups": self.config.get("dimension_filter_groups", []),
                "dataState": self.config.get("data_state", "final"),
            }
            response = (
                self.service.searchanalytics()
                .query(siteUrl=self.config.get("site_url"), body=request)
                .execute()
            )

        except Exception as err:
            # print(err)
            raise err

        return response

    def run_query(self):
        """Main Class Method"""
        # site_list = self.service.sites().list().execute()
        # print(site_list)

        for dimension in self.config["dimensions"]:
            for start_date, end_date in date_range_iterator(
                start_date=self.config["start_date"],
                end_date=self.config["end_date"],
                interval=self.config.get("interval", "day"),
                end_inclusive=True,
                time_format="%Y-%m-%d",
            ):
                dataset: List[Dict[str, Any]] = []
                query_dim_list: List[str] = list(dict.fromkeys(["date", dimension]))
                start_row: int = 0
                while True:
                    response = self._query_handler(
                        start_date=start_date,
                        end_date=end_date,
                        start_row=start_row,
                        query_dim_list=query_dim_list,
                    )

                    if response is None:
                        self.not_success()
                        print(
                            f"response is 'None' for date dimension row: "
                            f"'{start_date}' '{dimension}' '{start_row}'"
                        )
                        break
                    if "rows" not in response:
                        self.not_success()
                        print(
                            f"No more data for date dimension row: "
                            f"'{start_date}' '{dimension}' '{start_row}'"
                        )
                        break

                    self.is_success()
                    normalised_dataset = self._normalize(response, query_dim_list)
                    dataset.extend(normalised_dataset)
                    start_row += 1

                yield {
                    "date": start_date.replace("-", ""),
                    "dimension": dimension,
                    "data": dataset,
                }



"""MODULE FOR ONESIGNAL"""
# pylint: disable=arguments-differ, too-few-public-methods
import sys
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.base_readers import BaseReader
from sdc_dp_helpers.onesignal.onesignal_sdk import OneSignalHandlerFactory


class OneSignalReader(BaseReader):
    """reader for one signal"""

    def __init__(self, creds_filepath: str, config_filepath: str):
        self.creds: dict = load_file(creds_filepath)
        self.configs: dict = load_file(config_filepath)
        self.service = self._get_auth()

    def _get_auth(self):
        """class method to acquire the onesignal api sdk service"""
        handler_factory = OneSignalHandlerFactory()
        return handler_factory.get_endpoint_handler(
            creds=self.creds, configs=self.configs
        )

    def _query_handler(self):
        """make api call to endpoint"""
        yield from self.service.fetch_data()

    def run_query(self):
        """main loop for the reader"""
        results = self._query_handler()
        if not results:
            self.not_success()
            print(
                f"No data got for {self.configs['endpoint']} for {self.creds['app_id']}"
            )
            sys.exit()
        for result in results:
            self.is_success()
            yield {"date": result["date"], "data": result["data"]}



# pylint: disable=too-few-public-methods

"""
    CUSTOM WRITER CLASSES
"""
import re
import json

# from datetime import datetime

import boto3


class CustomFalconWriter:
    """ "
    Writer for falcon API
    Works with dictionary object from the readers
    """

    def __init__(self, bucket, file_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket = bucket
        self.file_path = file_path
        self.success = False

    def is_success(self):
        """Updates the Write Sucsess to True"""
        self.success = True

    def write_to_s3(self, payload: dict) -> None:
        """
        This pulls the yielded dataset from the GA reader in a manner
        that consumes the dataset of the given view_id and date,
        and writes it to s3 so that duplication does not occur.
        :param payload: This is a key value object that looks like:
                        {
                            "data": list(),
                            "date": string,
                            "networks": string
                        }
        """

        # confirm the payload keys are matching accurately with what is expected
        if not {"data", "date", "networks"}.issubset(set(payload.keys())):
            raise KeyError("Invalid payload expecting: date, data, networks")

        if not re.search(r"\d{4}\-\d{2}\-\d{2}", payload["date"]):
            raise ValueError(
                "Invalid date format for partitioning expected: 'YYYY-mm-dd'"
            )

        if not isinstance(payload["data"], list):
            raise TypeError("Invalid data passed: expected List[Dict, Dict]")

        _data: list = payload["data"]
        _date: str = payload["date"].replace("-", "")
        _networks: str = payload["networks"]

        write_path: str = f"{self.file_path}/{_networks}/{_date}.json"
        if _data:
            print(
                f"Writing data to s3://{self.bucket}/{write_path}  \
                partitioned by networks and date."
            )
            self.s3_resource.Object(self.bucket, write_path).put(Body=json.dumps(_data))
            self.is_success()



"""
sdc_dp_helpers.exceptions
~~~~~~~~~~~~~~~~~~~~
Exceptions used in sdc_dp_helpers.
"""


class SDCHelpersException(Exception):
    """A base class for sdc_dp_helpers' exceptions."""


class AuthenticationError(SDCHelpersException):
    """An error during authentication process."""


class QueryHandlerError(SDCHelpersException):
    """An exception raise when query handler fails."""


class DataNotFoundError(SDCHelpersException):
    """An error raised when we do not find data from the API Call."""


class RequestURITooLongError(SDCHelpersException):
    """Request-URI Too Long Error."""


class InvalidEndPointError(SDCHelpersException):
    """Invalid endpoint Error."""


class APICallError(SDCHelpersException):
    """Error for API Call that calls for retry."""


class TimeIntervalError(SDCHelpersException):
    """Exception Raised When you provide a wrong time interval value

    Args:
        SDCHelpersException (_type_): _description_
    """


class DateValueError(SDCHelpersException):
    """Exception Raised When you provide a wrong time interval value

    Args:
        SDCHelpersException (_type_): _description_
    """


class CadenceError(SDCHelpersException):
    """Exception Raised When you provide a wrong time cadence value

    Args:
        SDCHelpersException (_type_): _description_
    """


class TimeBucketError(SDCHelpersException):
    """Exception Raised When you provide a wrong time bucket value

    Args:
        SDCHelpersException (_type_): _description_
    """


class CadenceTimeBucketError(SDCHelpersException):
    """Exception Raised When you provide a wrong cadence or time bucket value

    Args:
        SDCHelpersException (_type_): _description_
    """



# pylint: disable=too-few-public-methods
# pylint: disable=no-member

"""
    CUSTOM WRITER CLASSES
"""
import json

import boto3


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, folder_path, profile_name=None):
        self.bucket = bucket
        self.folder_path = folder_path

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")

    def write_partition_to_s3(self, json_data):
        """
        Partitions data by date to reduce duplication in s3.
        So the plan is to ensure that if we:
            - pull from a specific date, we overwrite all that days data in the buckets.
            - this way, we ensure all data is original and in its most updated state.
        """

        date = json_data["date"].replace("-", "")
        dataset = json_data["data"]
        partition = json_data["partition"]
        venture = json_data["venture"]
        key_path = f"{self.folder_path}/{venture}/{date}/{date}_{partition}.json"
        print(f"Writing data to {key_path}.")
        self.s3_resource.Object(self.bucket, key_path).put(Body=json.dumps(dataset))



"""
    Utils functions
"""
import yaml


def init_config(file_name: str = "./example.yaml"):
    """
    Util method that creates an example config at the given file_name.
    :file_name: Full path to config YAML file.
    """
    config = {
        "blast": {"status": "sent", "start_date": "yesterday", "end_date": "yesterday"},
        "list": {"start_date": "yesterday", "end_date": "yesterday"},
        "send": {"start_date": "yesterday", "end_date": "yesterday"},
    }

    # write to file
    with open(file_name, "w") as file_:
        yaml.dump(config, file_)



"""
    CUSTOM READER CLASS
"""
# pylint: disable=too-few-public-methods,import-error,unused-import,redefined-outer-name
from typing import Dict, List, Any
from google.analytics.data_v1beta import BetaAnalyticsDataClient
from google.analytics.data_v1beta.types import DateRange
from google.analytics.data_v1beta.types import Dimension
from google.analytics.data_v1beta.types import Metric
from google.analytics.data_v1beta.types import RunReportRequest


from sdc_dp_helpers.api_utilities.date_managers import date_range, date_range_iterator
from sdc_dp_helpers.api_utilities.file_managers import load_file


class GAV4Reader:
    """
    GOOGLE ANALYTICS V4 READERS CLASS
    """

    def __init__(self, configs_file_path: str, service_account_file_path: str):
        self.configs = load_file(configs_file_path, fmt="yml")
        self.service_account_file_path = service_account_file_path
        self._client = self._get_client()
        self.dataset = []

    def _get_client(self):
        client = BetaAnalyticsDataClient().from_service_account_json(
            self.service_account_file_path
        )
        return client

    def _normalize(self, data, property_id: str) -> List[Dict[Any, Any]]:
        """Normalizes Data to Dictionary Format"""
        list_dataset = []
        dimension_headers = data.dimension_headers
        metric_headers = data.metric_headers

        for idx, row in enumerate(data.rows):
            row_data = {
                "property_id": property_id,
                "profile_name": self.configs["property_ids"][property_id],
            }

            for idx, dim_value_key in enumerate(row.dimension_values):
                row_data[dimension_headers[idx].name] = dim_value_key.value

            for idx, metric_value_key in enumerate(row.metric_values):
                row_data[metric_headers[idx].name] = metric_value_key.value

            list_dataset.append(row_data)
            # print(row_data)
        return list_dataset

    def _query_handler(self, property_id: str, start_date: str, end_date: str):
        """Runs a simple report on a Google Analytics 4 property."""
        # Explicitly use service account credentials by specifying
        # the private key file.
        # query = self.build_query(property_id,date)
        # request = RunReportRequest(**query)
        request = RunReportRequest(
            property=f"properties/{property_id}",
            dimensions=[Dimension(name=dim) for dim in self.configs["dimensions"]],
            metrics=[Metric(name=metric) for metric in self.configs["metrics"]],
            date_ranges=[DateRange(start_date=start_date, end_date=end_date)],
        )
        response = self._client.run_report(request)

        return response
        # [END analyticsdata_json_credentials_run_report]

    def run_query(self):
        """Controls the Flow of Query"""

        for property_id in self.configs["property_ids"]:
            for start_date, end_date in date_range_iterator(
                start_date=self.configs["start_date"],
                end_date=self.configs["end_date"],
                interval=self.configs["interval"],
                end_inclusive=True,
                time_format="%Y-%m-%d",
            ):
                payload = self._query_handler(
                    property_id=property_id, start_date=start_date, end_date=end_date
                )
                if payload:
                    dataset: List[Dict] = self._normalize(payload, property_id)
                    yield {
                        "date": start_date,
                        "property_id": property_id,
                        "data": dataset,
                    }
                    self.dataset = dataset

    # def build_query(self, property_id, date):
    #     query = {
    #             "property":f"properties/{property_id}",
    #             "dimensions":[Dimension(name=dim) for dim in self.configs["dimensions"]],
    #             "metrics":[Metric(name=metric) for metric in self.configs.get("metrics", [])],
    #             "date_ranges":[DateRange(start_date=date, end_date=date)]
    #     }
    #     return query



from ftplib import FTP
from io import BytesIO
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler


class FTPReader:
    """
    Custom FTP Reader
    """

    def __init__(self, creds_file, config_file):
        self._creds: dict = load_file(creds_file, "yml")

        self.config = load_file(config_file)

    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    def _query_handler(self) -> str:
        """
        Fetch XML from url
        """

        ftp = FTP(host=str(self.config.get("hostname")),
                  user=str(self._creds.get("user")),
                  passwd=str(self._creds.get("password")))

        ftp.encoding = "utf-8"
        read_io = BytesIO()

        try:
            ftp.retrbinary(f"RETR {self.config.get('filename')}", read_io.write)

        except Exception as err:
            raise err

        return read_io.getvalue().decode('utf8')

    def run_query(self) -> list:

        data = self._query_handler()

        # Split rows by newline
        data = data.split('\n')

        # Header is the first line
        headers = data[0].split('\t')

        # Last item is a timestamp
        lines = data[1:-1]

        # Split columns by tab
        list_of_lines = []
        for line in lines:
            list_of_lines.append(line.split('\t'))

        # Create list of dictionaries where the headers are the keys
        result = [dict(zip(headers, values)) for values in list_of_lines]

        return result



"""MODULE TO HELP WITH PROCESSING FILES"""
# pylint: disable=unused-argument, inconsistent-return-statements
from typing import Union, Dict, Any, Generator
from pathlib import Path
import json
from io import BytesIO
from io import TextIOWrapper
from csv import DictReader
from zipfile import ZipFile
from gzip import GzipFile
import yaml


def load_file(file_location: str, fmt: Union[str, None] = None) -> Dict[Any, Any]:
    """
    Gathers file data from json or yaml.
    """
    config: dict = {}
    if file_location.strip().rsplit(".", maxsplit=1)[-1] not in ["json", "yml", "yaml"]:
        raise TypeError("Wrong file type provided! Expecting only json and yaml files")

    file_location = str(file_location).strip()
    if file_location.endswith("yml"):
        with open(file_location, mode="r", encoding="utf8") as yaml_file:
            config = yaml.safe_load(yaml_file)
    if file_location.endswith("json"):
        with open(file_location, mode="r", encoding="utf8") as json_file:
            config = json.load(json_file)

    return config


def local_json_writer(
    payload: dict,
    filename: str,
    root_dir: str,
    folder_path: Union[str, None] = None,
    mode: str = "w",
) -> None:
    """WRITE DATA TO JSON OBJECT"""

    file_path = root_dir
    if folder_path:
        file_path = f"{root_dir}/{folder_path}"
        # CREATE FILE PATH IF NOT EXISTS
    Path(file_path).mkdir(parents=True, exist_ok=True)

    if filename is None or str(filename).strip() == "":
        raise ValueError("please proivde a destination filename")

    filename = f"{file_path}/{filename}.json"

    with open(f"{filename}", mode=mode, encoding="utf8") as dest_file:
        json.dump(payload, dest_file, indent=4)
        print(f"Done writting data to {filename}")

def parse_zip_to_csv(
    response, file_type: Union[str, None] = None
) -> Generator[Dict, None, None]:
    """Parse the zip file from api call"""
    if file_type == "zip":
        with ZipFile(BytesIO(response)) as zipfile:
            for file in zipfile.filelist:
                with zipfile.open(file) as file_contents:
                    dict_reader = DictReader(
                        TextIOWrapper(file_contents, encoding="utf-8")
                    )
                    yield from dict_reader
    elif file_type == "gzip":
        byte_object = BytesIO(response)
        byte_object.seek(0)
        with GzipFile(fileobj=byte_object, mode="r") as zipfile:
            dict_reader = DictReader(TextIOWrapper(zipfile, encoding="utf-8"))
            yield from dict_reader



# pylint: disable=too-few-public-methods,arguments-differ,import-error,too-many-arguments
""" Falcon Reader SDK"""
from abc import abstractmethod
from typing import List, Dict
import os
from datetime import datetime
import requests

from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler


class RequestHandler:
    """Interface for  API Call method"""

    def __init__(self, session, creds):
        self.session: requests.Session = session
        self.base_url: str = "https://api.falcon.io/"
        self._creds = creds

    @abstractmethod
    def make_api_call(self, **kwargs):
        """Make API Call"""
        raise NotImplementedError


class FalconAPICall:
    """Class for Making API Call"""

    def __init__(self, session, creds, config):
        self.session = session
        self._creds = creds
        self.config = config
        self.channel_ids = self.get_channel_ids()
        self.start = 0
        self.curr_channels = []

    def get_channel_ids(self):
        """This gets channel ids"""
        channel_id_methods = {"v1": ChannelIdsV1, "v2": ChannelIdsV2}
        version = self.config["version"]
        if version not in channel_id_methods:
            raise KeyError(
                f"unkown value for 'version' in configs, expecting 'v1' or 'v2'. Got {version}"
            )
        channel_ids_caller = channel_id_methods[version](
            session=self.session, creds=self._creds
        )
        channel_ids = channel_ids_caller.make_api_call()

        return channel_ids

    def get_data(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        """Gets the API Call Handler to Use"""
        endpoint_handlers = {
            "channel_insights": ChannelInsights,
            "content_insights": ContentInsights,
            "published_posts": PublishedPosts,
        }
        endpoint_name = self.config["endpoint_name"]
        if endpoint_name not in endpoint_handlers:
            raise KeyError(
                f"""
                unkown endpoint name in configs, expecting 'channel_insights',
                'content_insights' or 'published_posts'. Got {endpoint_name}
                """
            )
        endpoint_caller = endpoint_handlers[endpoint_name](
            session=self.session, creds=self._creds, config=self.config
        )
        data = endpoint_caller.get_dataset(start_date, end_date, network, channel_ids)

        return data


class ChannelIdsV1(RequestHandler):
    """Class for v1 channel ids"""

    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def make_api_call(self, **kwargs):
        """
        Gather all available channel ids v1.
        """
        print("GET v1: channel ids.")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        response = self.session.get(url=url, params={"limit": "9999"})

        response_data = response.json()
        if response.status_code == 200:
            channel_ids = set()
            for item in response_data.get("items", []):
                channel_ids.add(item["id"])

            return list(channel_ids)

        raise ConnectionError(
            f"Falcon API failed to return channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )


class ChannelIdsV2(RequestHandler):
    """Class for v2 channel ids"""

    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def make_api_call(self, **kwargs):
        """
        Gather all available channel ids v2.
        """

        print("GET v2: channel ids.")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        response = self.session.get(url=url, params={"limit": "9999"})

        response_data = response.json()
        if response.status_code == 200:
            channel_ids = {}
            for item in response_data.get("items", []):
                channel_ids[item["uuid"]] = item["name"]
            return channel_ids

        raise ConnectionError(
            f"Falcon API failed to return channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )


class ContentIdByChannelId(RequestHandler):
    """Class to get Content Ids by Channel id"""

    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(
        self, start_date: str, end_date: str, network: str, channel_ids: Dict[str, str]
    ) -> dict:
        print("GET: content ids by channel id.")
        date_filters = f"since={start_date}&until={end_date}"
        endpoint_url = (
            f"publish/items?apikey={self._creds['api_key']}"
            f"&statuses=published"
            f"&networks={network}"
            f"&{date_filters}"
        )
        content_ids_by_channel_id = {}
        while endpoint_url is not None:
            url = f"{self.base_url}{endpoint_url}"
            response = self.session.get(url=url)
            response_data = response.json()
            if response.status_code != 200:
                raise ConnectionError(
                    f"Falcon API failed to return content ids. "
                    f"Status code: {response.status_code}, Reason: {response.reason}."
                )
            for item in response_data.get("items"):
                content_id = item.get("id")
                channel_id = item.get("channels")
                if channel_id is not None:
                    channel_id = channel_id[0]
                    if network == "linkedIn":
                        channel_id = channel_id.replace("-", "")

                    if channel_id in channel_ids:
                        content_ids_by_channel_id.setdefault(channel_id, []).append(
                            content_id
                        )
            endpoint_url = response_data.get("next", {"href": None}).get("href")
        return content_ids_by_channel_id


class PublishedPostsByChannelId(RequestHandler):
    """Gets the published posts by channel id"""

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def make_api_call(
        self,
        start_date: str,
        end_date: str,
        network: str,
        channel_ids: list,
        statuses: str,
        limit: int,
        index: int,
    ) -> list:
        """Gets the published posts by channel id
        SEE: https://falconio.docs.apiary.io/
        #reference/content-api/get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: list of dictionaries
        """

        dataset: list = []
        endpoint_url: str = f"publish/items?apikey={self._creds['api_key']}"
        while endpoint_url:
            print(
                f"INFO: channel id index: {index}, "
                f"channel id: {channel_ids},  date: {start_date}, offset: {len(dataset)}."
            )
            response = self.session.get(
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                params={
                    "channels": channel_ids,
                    "since": start_date,
                    "until": end_date,
                    "networks": network,
                    "statuses": statuses,
                    "limit": limit,
                },
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if response.status_code == 200:
                results: dict = response.json()
                items_data: list = results.get("items", [])
                dataset.extend(items_data)

                endpoint_url = results.get("next", {"href": None}).get("href")

                if len(items_data) == 0 or endpoint_url is None:
                    break

            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel id: {channel_ids}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset


class ContentInsightsRequestId(RequestHandler):
    """This class gets Content Insights Request Id"""

    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(
        self,
        start_date: str,
        end_date: str,
        metric_ids: List[str],
        content_ids_by_channel_id: Dict[str, List],
    ) -> Dict[str, str]:
        print("GET: insights request id.")
        endpoint_url = f"measure/v2/insights/content?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        body = {
            "since": start_date,
            "until": end_date,
            "metricIds": metric_ids,
            "channels": [],
        }
        for key, value in content_ids_by_channel_id.items():
            body["channels"].append({"id": key, "contentIds": value})
        response = self.session.post(
            url=url, headers={"Content-Type": "application/json"}, json=body
        )
        if response.status_code == 200:
            response_data = response.json()
            insights_request_id = response_data["insightsRequestId"]
            return insights_request_id

        raise ConnectionError(
            f"Falcon API failed to return content ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )


class ChannelInsightsRequestId(RequestHandler):
    """This class gets Channel Insights Request Id"""

    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(
        self,
        start_date: str,
        end_date: str,
        metric_ids: List[str],
        channel_ids: List[str],
    ) -> Dict[str, str]:
        print("GET: insights request id.")
        start_date = datetime.strftime(
            datetime.strptime(start_date, "%Y-%m-%dT%H:%M:%SZ"), "%Y-%m-%d"
        )
        end_date = start_date
        endpoint_url = f"measure/v2/insights/channel?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        body = {
            "since": start_date,
            "until": end_date,
            "metricIds": metric_ids,
            "channelIds": channel_ids,
        }
        response = self.session.post(
            url=url, headers={"Content-Type": "application/json"}, json=body
        )
        if response.status_code == 200:
            response_data = response.json()
            insights_request_id = response_data["insightsRequestId"]
            return insights_request_id

        raise ConnectionError(
            f"Falcon API failed to return content ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )


class Insights(RequestHandler):
    """This class gets insights data"""

    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(self, insights_request_id: Dict[str, str] = None):
        print("GET: insights.")
        endpoint_url = (
            f"measure/v2/insights/{insights_request_id}?apikey={self._creds['api_key']}"
        )
        url = f"{self.base_url}{endpoint_url}"
        response = self.session.get(url=url)

        response_data = response.json()
        if response.status_code == 200:
            if response_data["status"] == "READY":
                return response_data
            if response_data["status"] == "EXPIRED":
                return None
        raise ConnectionError(
            f"Falcon API failed to return content ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )


class ContentInsights(FalconAPICall):
    """This class  gets content insights and normalises the data"""

    @staticmethod
    def _normalize_data(dataset: dict, network: str, channel_ids: dict) -> List[Dict]:
        temp = {}

        for metric, data_items in dataset.items():
            if not data_items:
                # if we have nothing in value
                continue
            for data in data_items:
                data[metric] = data.pop("value")
                data["network"] = network
                data["brand"] = channel_ids.get(data["channelId"], None)
                check_point = tuple(map(data.get, ["channelId", "date", "contentId"]))
                if not temp.get(check_point):
                    temp[check_point] = data
                else:
                    missing_keys = set(data).difference(set(temp[check_point]))
                    # flatten
                    for key in missing_keys:
                        temp[check_point].update({key: data[key]})
        normalized_data = list(temp.values())
        return normalized_data

    def get_dataset(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        """Fucntion handles content insights query"""
        channel_steps = self.config.get("channel_steps", 10)
        metric_ids = self.config["metrics"]
        content_ids_by_channel_id = ContentIdByChannelId(
            creds=self._creds, session=self.session
        ).make_api_call(
            start_date=start_date,
            end_date=end_date,
            network=network,
            channel_ids=channel_ids,
        )
        channels = list(content_ids_by_channel_id.keys())
        channel_id_length = len(channels)
        dataset = []
        data = []
        for _ in range(0, channel_id_length, channel_steps):
            self.curr_channels = channels[self.start : self.start + channel_steps]
            request_content_ids_by_channel_id = {
                key: value
                for key, value in content_ids_by_channel_id.items()
                if key in self.curr_channels
            }

            if len(request_content_ids_by_channel_id) > 0:
                insights_request_id = ContentInsightsRequestId(
                    creds=self._creds, session=self.session
                ).make_api_call(
                    start_date=start_date,
                    end_date=end_date,
                    metric_ids=metric_ids,
                    content_ids_by_channel_id=request_content_ids_by_channel_id,
                )
                results = Insights(
                    creds=self._creds, session=self.session
                ).make_api_call(insights_request_id)
                if results is None:
                    insights_request_id = ContentInsightsRequestId(
                        creds=self._creds, session=self.session
                    ).make_api_call(
                        start_date=start_date,
                        end_date=end_date,
                        metric_ids=metric_ids,
                        content_ids_by_channel_id=request_content_ids_by_channel_id,
                    )
                    results = Insights(
                        creds=self._creds, session=self.session
                    ).make_api_call(insights_request_id)
                data = results["data"]["insights"]
                dataset.extend(self._normalize_data(data, network, channel_ids))
            self.start = self.start + channel_steps
        return dataset


class ChannelInsights(FalconAPICall):
    """This class  gets channel insights and normalises the data"""

    @staticmethod
    def _normalize_data(dataset: dict, network: str, channel_ids: dict) -> List[Dict]:
        templist = []

        for metric, data_items in dataset.items():
            if not data_items:
                # if we have nothing in value
                continue
            for data in data_items:
                data[metric] = data.pop("value")
                data["network"] = network
                data["brand"] = channel_ids.get(data["channelId"], None)
                templist.append(data)
        normalized_data = templist

        return normalized_data

    def get_dataset(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        """Fucntion handles channel insights query"""
        channels = list(channel_ids.keys())
        channel_id_length = len(channels)
        channel_steps = self.config.get("channel_steps", 10)
        metric_ids = self.config["metrics"]
        dataset = []
        data = []
        for _ in range(0, channel_id_length, channel_steps):
            self.curr_channels = channels[self.start : self.start + channel_steps]
            insights_request_id = ChannelInsightsRequestId(
                creds=self._creds, session=self.session
            ).make_api_call(
                start_date=start_date,
                end_date=end_date,
                metric_ids=metric_ids,
                channel_ids=self.curr_channels,
            )
            results = Insights(creds=self._creds, session=self.session).make_api_call(
                insights_request_id
            )
            if results is None:
                insights_request_id = ChannelInsightsRequestId(
                    creds=self._creds, session=self.session
                ).make_api_call(
                    start_date=start_date,
                    end_date=end_date,
                    metric_ids=metric_ids,
                    channel_ids=self.curr_channels,
                )
                results = Insights(
                    creds=self._creds, session=self.session
                ).make_api_call(insights_request_id)
            data = results["data"]["insights"]
            dataset.extend(self._normalize_data(data, network, channel_ids))
            self.start = self.start + channel_steps
        return dataset


class PublishedPosts(FalconAPICall):
    """This class gets published posts data"""

    def get_dataset(
        self, start_date: str, end_date: str, network: str, channel_ids: list
    ) -> dict:
        """Fucntion handles published posts query"""
        print(f"Attempting to gather metrics from {len(channel_ids)} channel ids.")
        statuses = self.config.get("statuses", "published")
        limit = self.config.get("limit", 2000)
        dataset = []
        for channel_id in channel_ids:
            self.curr_channels = [channel_id]
            results = PublishedPostsByChannelId(
                creds=self._creds, session=self.session
            ).make_api_call(
                start_date=start_date,
                end_date=end_date,
                network=network,
                channel_ids=[channel_id],
                statuses=statuses,
                limit=limit,
                index=self.start,
            )
            print(len(results))
            dataset.extend(results)
            self.start += 1
        return dataset



"""BASE READER FOR ALL READERS"""

from abc import ABC, abstractmethod
from typing import Union, Generator, Dict, List, Any
import requests
from requests.adapters import HTTPAdapter, Retry


class BaseReader(ABC):
    """Base Reader Class Guiding All Readers"""

    success: List[bool] = []

    def _get_auth(self):
        """Get Authorization use this method to authorize"""
        return True


    @abstractmethod
    def _query_handler(self, *args, **kwargs):
        """Handles the Query Call"""
        raise NotImplementedError

    @abstractmethod
    def run_query(
        self,
    ) -> Union[
        Generator[Dict[List[Dict[Any, Any]], Any], None, None],
        Dict[List[Dict[Any, Any]], Any],
    ]:
        """Calls the _query_handler method and handles iteration"""
        raise NotImplementedError

    def is_success(self):
        """Records the success status of the api call"""
        self.success.append(True)

    def not_success(self):
        """Records the failed status of the API call"""
        self.success.append(False)


class APIRequestsSessionsReader(BaseReader):
    """Base Requests Sessions
    https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
    """

    sessions: requests.Session = requests.Session()
    session_config: Dict[str, Union[int, bool, str, float]]

    def get_session(self):
        """Builds sessions with Retry Backed In"""
        retries = Retry(
            total=self.session_config.get("total", 3),
            backoff_factor=self.session_config.get("backoff_factor", 0.1),
            status_forcelist=self.session_config.get(
                "status_forcelist", [500, 502, 503, 504]
            ),
            read=self.session_config.get("read", 3),
            redirect=self.session_config.get("redirect", 3),
            raise_on_redirect=self.session_config.get("raise_on_redirect", True),
            raise_on_status=self.session_config.get("raise_on_status", True),
        )
        for protocol in ["http://", "https://"]:
            self.sessions.mount(protocol, HTTPAdapter(max_retries=retries))

        return self.sessions



"""UTILITIES TO HANDLE DATE VALUES"""

# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
import re
from typing import Tuple, Generator, Union
from dateutil.relativedelta import relativedelta
import pandas as pd
from sdc_dp_helpers.api_utilities.exceptions import (
    TimeIntervalError,
    TimeBucketError,
    CadenceTimeBucketError,
    DateValueError,
)


class DatePhraseHandler(ABC):
    """Handles Phrases In the Input Date"""

    date_value: Union[str, None] = None
    time_bucket: Union[str, None] = None
    cadence: Union[int, None] = None

    def __init__(self, date_value: str):
        self.date_value = date_value

    @abstractmethod
    def phrase_to_date(self) -> str:
        """Processes the date_value and translates it to a datetime value
        Raises:
            NotImplementedError: _description_
        Returns:
            str: _description_
        """
        raise NotImplementedError


class BaseDateInterval(ABC):
    """Get Start of the Date Range"""

    def __init__(self, time_bucket: str, cadence: int):
        self.time_bucket = time_bucket
        self.cadence = cadence

    @abstractmethod
    def get_start_date(self, start_date: datetime) -> datetime:
        """Gets the start date of the range provided the input startdate
        Args:
            start_date (datetime): should be a datetime value
        Raises:
            NotImplementedError: _description_
        Returns:
            datetime: the processed start date
        """
        raise NotImplementedError

    @abstractmethod
    def add_interval(self, date_value: datetime, cadence: int) -> datetime:
        """Add the time interval to the current start date to get the end date
        Args:
            date_value (datetime): datetime value that is the start of the period
            cadence (int): time interval to add to the start date tp the end date
        Raises:
            NotImplementedError: raise error of NotImplementedError
        Returns:
            datetime: the end date value not inclusive
        """
        raise NotImplementedError

    @abstractmethod
    def subtract_interval(self, date_value: datetime, cadence: int) -> datetime:
        """Subtract the time interval to the current start date to get the end date
        Args:
            date_value (datetime): datetime value that is the start of the period
            cadence (int): time interval to add to the start date tp the end date
        Raises:
            NotImplementedError: raise error of NotImplementedError
        Returns:
            datetime: the end date value not inclusive
        """
        raise NotImplementedError


class YearInterval(BaseDateInterval):
    """Class ti Get Yearly Cadence Start Date"""

    def __init__(self, time_bucket: str, cadence: int):
        super().__init__(time_bucket=time_bucket, cadence=cadence)

    def get_start_date(self, start_date: datetime) -> datetime:
        """Handles Yearly Buckets"""
        if self.time_bucket in ["yearly", "last_year", "this_year"]:
            start_date = start_date.replace(month=1, day=1)
        return start_date

    def add_interval(self, date_value: datetime, cadence: int):
        """Return a date that's `years` years after the date (or datetime)
        object `date_value`. Return the same calendar date (month and day) in the
        destination year, if it exists, otherwise use the following day
        (thus changing February 29 to March 1).
        """
        next_date = date_value + relativedelta(years=cadence)
        return next_date

    def subtract_interval(self, date_value: datetime, cadence: int):
        next_date = date_value - relativedelta(years=cadence)
        return next_date


class MonthInterval(BaseDateInterval):
    """Gets the Monthly Cadence Start Date"""

    def __init__(self, time_bucket: str, cadence: int):
        super().__init__(time_bucket=time_bucket, cadence=cadence)

    def get_start_date(self, start_date: datetime) -> datetime:
        """Handles Monthly Buckets"""
        if self.time_bucket in ["monthly", "month", "last_month", "this_month"]:
            start_date = start_date.replace(day=1)
        return start_date

    def add_interval(self, date_value: datetime, cadence: int):
        next_date = date_value + relativedelta(months=cadence)
        return next_date

    def subtract_interval(self, date_value: datetime, cadence: int):
        next_date = date_value - relativedelta(months=cadence)
        return next_date


class WeekInterval(BaseDateInterval):
    """Get the Weekly Cadence Start Date"""

    def __init__(self, time_bucket: str, cadence: int):
        super().__init__(time_bucket=time_bucket, cadence=cadence)

    def get_start_date(self, start_date: datetime) -> datetime:
        """Handles Weekly Buckets"""
        if self.time_bucket in ["weekly", "last_week", "this_week"]:
            if datetime.strftime(start_date, "%A") == "Sunday":
                return start_date
            start_date = start_date - timedelta(days=start_date.weekday() + 1)
        return start_date

    def add_interval(self, date_value: datetime, cadence: int):
        next_date = date_value + relativedelta(weeks=cadence)
        return next_date

    def subtract_interval(self, date_value: datetime, cadence: int):
        next_date = date_value - relativedelta(weeks=cadence)
        return next_date


class DayInterval(BaseDateInterval):
    """Get the Daily Cadence Start Date"""

    def __init__(self, time_bucket: str, cadence: int):
        super().__init__(time_bucket=time_bucket, cadence=cadence)

    def get_start_date(self, start_date: datetime) -> datetime:
        """Handles Daily Buckets"""
        return start_date

    def add_interval(self, date_value: datetime, cadence: int):
        next_date = date_value + relativedelta(days=cadence)
        return next_date

    def subtract_interval(self, date_value: datetime, cadence: int):
        next_date = date_value - relativedelta(days=cadence)
        return next_date


class PastDatePhraseHandler(DatePhraseHandler):
    """Handles Past Date Phrases"""

    def phrase_to_date(self) -> Tuple[int, str]:
        """Processes the date value to return self.cadence and bucket
        Args:
            date_value (str): string representing date
        Raises:
            ValueError: if passed a wrong and unexpected date value
        Returns:
            Tuple[int, str]: self.cadence and bucket(day, week, month, year)
        """
        pattern = re.compile(r"[a-z]+")
        if not isinstance(self.date_value, str) or not pattern.search(
            self.date_value.strip().lower()
        ):
            raise TypeError(
                f"wrong date_string provided, expecting a string but got: {self.date_value}"
            )
        date_value = self.date_value.lower().strip()
        if date_value in ["yesterday", "last_week", "last_month", "last_year"]:
            self.cadence, self.time_bucket = 1, date_value
        if date_value in ["today", "this_week", "this_year", "this_month"]:
            self.cadence, self.time_bucket = 0, date_value
        if re.search(r"\_", date_value) and self.time_bucket is None:
            if len(date_value.split("_")) != 3 or date_value.split("_")[2] != "ago":
                raise DateValueError(f"wrong date string provided: {date_value}")
            if len(date_value.split("_")) == 3:
                self.cadence, self.time_bucket = date_value.split("_")[:-1]
                if self.time_bucket.endswith("s"):
                    self.time_bucket = self.time_bucket.replace("s", "")
        if self.cadence is None or self.time_bucket is None:
            raise CadenceTimeBucketError(
                f"cadence or time_bucket is None, wrong date_string provided: '{date_value}'"
            )
        # print(self.cadence, self.time_bucket)
        self.cadence, self.time_bucket = int(self.cadence), self.time_bucket.strip()
        return self.cadence, self.time_bucket


class IntervalDatePhraseHandler(DatePhraseHandler):
    """Class for handling interval date string"""

    def phrase_to_date(self) -> Tuple[int, str]:
        """Processes the interval variable to get the cadence and the time_bucket
        Raises:
            TypeError: is the interval value p[rovided is not a string]
            ValueError: if the interval can be split more than 2 times
            ValueError: if time_bucket and cadence end up to be None
        Returns:
            Tuple[int, str]: cadence that is an int and time_bucket
        """
        if not isinstance(self.date_value, str) or not re.search(
            r"[a-z]+", str(self.date_value).lower()
        ):
            raise TypeError(
                f"wrong interval provided, expecting a string but got: {self.date_value}"
            )
        date_value = self.date_value.lower().strip()
        if date_value in ["day", "yearly", "weekly", "monthly"]:
            self.cadence, self.time_bucket = 1, date_value
        if len(date_value.split("_")) > 2 and self.time_bucket is None:
            raise TimeIntervalError(
                f"invalid value for interval provided: {date_value}"
            )
        if len(self.date_value.replace(" ", "_").split("_")) == 2:
            self.cadence, self.time_bucket = date_value.split("_")
        if self.cadence is None or self.time_bucket is None:
            raise CadenceTimeBucketError(
                f"cadence or time_bucket is None, wrong interval provided: '{date_value}'"
            )
        # print(self.cadence, self.time_bucket)
        self.cadence, self.time_bucket = int(self.cadence), self.time_bucket.strip()
        return self.cadence, self.time_bucket


class DateHandlerFactory:
    """A class that handles provided date to translate it to valie datetime object"""

    def get_date_handler(self, time_bucket: str, cadence: int) -> BaseDateInterval:
        """Factory method to determine what time bucket we are going to process
        Args:
            time_bucket (str): either day, week, month, year
            cadence (int): cadence to be used to subtract or add date
        Raises:
            TimeBucketError: If the time bucket provided is not catered for
        Returns:
            BaseDateInterval: Returns and instance of BaseDateInterval
        """
        root_bucket = None
        time_bucket_mapping = {
            "day": ["day", "yesterday", "today"],
            "week": ["week", "weekly", "last_week", "this_week"],
            "month": ["month", "monthly", "last_month", "this_month"],
            "year": ["year", "yearly", "last_year", "this_year"],
        }
        for key, val in time_bucket_mapping.items():
            if time_bucket in val:
                root_bucket = key
        if root_bucket is None:
            raise TimeBucketError(f"wrong time bucket in the provided: '{time_bucket}'")
        date_handlers = {
            "day": DayInterval,
            "week": WeekInterval,
            "month": MonthInterval,
            "year": YearInterval,
        }
        return date_handlers[root_bucket](time_bucket=time_bucket, cadence=cadence)


def date_string_handler(date_string: str) -> datetime:
    """Processes the Date and Returns the date value to work with
    Args:
        date_string (str): used to determine the date '2022-11-14', 2_days_ago, yeaterday, today
    Returns:
        datetime: datetime value
    """
    date_string = str(date_string).strip()
    string_search = re.search(r"\d{4}\-\d{2}\-\d{2}", date_string)
    if string_search:
        date_part = string_search.group()
        return datetime.strptime(date_part, "%Y-%m-%d")
    if date_string == "tomorrow":
        return datetime.strptime(
            (datetime.now() + timedelta(days=1)).date().isoformat(), "%Y-%m-%d"
        )
    phrase_handler = PastDatePhraseHandler(date_value=date_string)
    cadence, time_bucket = phrase_handler.phrase_to_date()
    handler = DateHandlerFactory().get_date_handler(
        time_bucket=time_bucket, cadence=cadence
    )
    start_date = datetime.now()
    start_date = handler.get_start_date(start_date=start_date)
    date_value = handler.subtract_interval(date_value=start_date, cadence=cadence)
    return date_value


def date_range_iterator(
    start_date: str,
    end_date: str,
    interval: str,
    end_inclusive: bool = False,
    time_format: str = "%Y-%m-%d",
) -> Generator[Tuple[str, str], None, None]:
    """Processes the Date and Returns the date value to work with
    Args:
        start_date (str): date string (2022-11-14 or the allowed date string values)
        end_date (str): date string (2022-11-14 or the allowed date string values)
        interval (str): string 1_month, 1_day, yearly, monthly, weekly
        end_inclusive (bool): whether the yielded end for period/interval is inclusive or not.
        time_format (str, optional): The format of the returned date value. Defaults to "%Y-%m-%d".
    Yields:
        Generator[Tuple[str, str], None, None]: start and end (exclusive) for each time interval.
    """
    phrase_handler = IntervalDatePhraseHandler(date_value=interval)
    cadence, time_bucket = phrase_handler.phrase_to_date()
    handler = DateHandlerFactory().get_date_handler(
        time_bucket=time_bucket, cadence=cadence
    )
    startdate = date_string_handler(start_date)
    enddate = date_string_handler(end_date)
    startdate = handler.get_start_date(start_date=startdate)
    next_end = handler.add_interval(date_value=startdate, cadence=cadence)
    while startdate <= enddate:
        end = next_end
        if end_inclusive:
            end = next_end - timedelta(days=1)
        yield datetime.strftime(startdate, time_format), datetime.strftime(
            end, time_format
        )
        startdate = next_end
        next_end = handler.add_interval(date_value=startdate, cadence=cadence)


def date_handler(date_string, date_format="%Y-%m-%d"):
    """
    Takes a date string or phrase and returns the valid date in
    the specified format.
    :date_string: A phrase like "3_days_ago", "yesterday" or "today"
                     or a date string such as "2021-01-01".
    :date_format: The returned formatting of the date.
    """
    try:
        if date_string == "today":
            return datetime.now().strftime(date_format)
        if date_string == "yesterday":
            return (datetime.now() - timedelta(days=1)).strftime(date_format)
        if "_days_ago" in date_string:
            return phrase_to_date(phrase=date_string, date_format=date_format)
        return date_string
    except ValueError as err:
        raise ValueError(
            "StartDate requires a valid input "
            'such as "today", "yesterday" or "<int>_days_ago".'
        ) from err


def phrase_to_date(phrase, date_format="%Y-%m-%d"):
    """
    Takes a typical phrase such as "3_days_ago" and returns a
    date based on that phrase.
    :phrase: str. Something like 3_days_ago or 25_days_ago etc.
    :return: '%Y-%m-%d'
    """
    try:
        date_delta = int(phrase.split("_")[0])
        return (datetime.now() - timedelta(days=date_delta)).strftime(date_format)
    except ValueError as err:
        raise ValueError(
            f"Phrasing for date: {phrase} is not valid, "
            f"try something like: 3_days_ago"
        ) from err


def date_range(start_date, end_date, delta=timedelta(days=1)):
    """
    The range is inclusive, so both start_date and end_date will be returned.
    :start_date: The datetime object representing the first day in the range.
    :end_date: The datetime object representing the second day in the range.
    :delta: A datetime.timedelta instance, specifying the step interval. Defaults to one day.
    Yields:
        Each datetime object in the range.
    """
    start_date = date_handler(date_string=start_date)
    end_date = date_handler(date_string=end_date)
    start_date = datetime.strptime(start_date, "%Y-%m-%d")
    end_date = datetime.strptime(end_date, "%Y-%m-%d")
    current_date = start_date
    while current_date <= end_date:
        yield current_date.strftime("%Y-%m-%d")
        current_date += delta


# pylint: disable=invalid-name
def filter_data_by_dates(
    start_date, end_date, date_field, data_frame: pd.DataFrame, date_fmt="%Y-%m-%d"
):
    """
    Takes a data frame and filters the data by given date field and
    date scopes that can be phrases.
    If no scope is added, imply current time is expected.
    """
    _now = datetime.now().strftime(date_fmt)
    if start_date is None:
        sd = _now
    else:
        sd = date_handler(start_date, date_fmt)
    if end_date is None:
        ed = _now
    else:
        ed = date_handler(end_date, date_fmt)
    print(f"Gathering data between {sd} and {ed}.")
    # filter data by given date field
    if sd is not None and ed is not None:
        data_frame["tmp_date"] = pd.to_datetime(data_frame[date_field])
        data_frame["tmp_date"] = data_frame["tmp_date"].dt.strftime(date_fmt)
        data_frame = data_frame[
            (
                data_frame["tmp_date"]
                >= datetime.strptime(sd, date_fmt).strftime(date_fmt)
            )
            & (
                data_frame["tmp_date"]
                <= datetime.strptime(ed, date_fmt).strftime(date_fmt)
            )
        ]
        data_frame = data_frame.drop("tmp_date", axis="columns")
    if len(data_frame.index) > 0:
        return data_frame
    print(f"No data for given date filter: {sd} to {ed}.")
    return None



# pylint: disable=invalid-name,line-too-long,pointless-string-statement
"""
    Set up the sdc_helpers package
"""
from distutils.command.sdist import sdist as _sdist
from setuptools import setup


class sdistzip(_sdist):
    """Override sdist to always produce .zip archive"""

    def initialize_options(self):
        _sdist.initialize_options(self)
        self.formats = "zip"


setup(
    name="sdc_dp_helpers",
    packages=[
        "sdc_dp_helpers",
        "sdc_dp_helpers.google_analytics",
        "sdc_dp_helpers.google_analytics_v4",
        "sdc_dp_helpers.google_search_console",
        "sdc_dp_helpers.sailthru",
        "sdc_dp_helpers.azure",
        "sdc_dp_helpers.api_utilities",
        "sdc_dp_helpers.falcon",
        "sdc_dp_helpers.pyspark",
        "sdc_dp_helpers.onesignal",
        "sdc_dp_helpers.xero",
        "sdc_dp_helpers.zoho",
        "sdc_dp_helpers.google_ads",
        "sdc_dp_helpers.facebook",
        "sdc_dp_helpers.google_big_query",
        "sdc_dp_helpers.xml",
        "sdc_dp_helpers.google_knowledge_graph",
        "sdc_dp_helpers.google_postmaster_tools",
        "sdc_dp_helpers.zoho_recruit",
        "sdc_dp_helpers.ftp",
    ],
    install_requires=[
        "boto3>=1.17.112",
        "google-api-python-client",
        "httplib2>=0.19.1",
        "oauth2client",
        "numpy",
        "pandas",
        "pyOpenSSL",
        "python-interface",
        "sailthru-client",
        "azure-storage-blob",
        "sdc-helpers==1.6.2",
        "requests>=2.25.1",
        "zcrmsdk==2.0.10",
        "mysql",
        "mysql-connector>=2.2.9",
        "googleads==28.0.0",
        "google-ads==12.0.0",
        "protobuf>=3.20.2",
        "pyxero>=0.9.3",
        "oauth2",
        "python-dateutil>=2.8.1",
        "google-cloud-bigquery",
        "google",
        "google-analytics-data",
        "xmltodict",
    ],
    extras_require={"pyspark": ["pyspark"]},
    cmdclass={"sdist_zip": sdistzip},
    description="A module for developing data pipelines from external api's and on ETL like services",
    version="1.3.35",
    url="http://github.com/RingierIMU/sdc-dataPipeline-helpers",
    author="Ringier South Africa",
    author_email="tools@ringier.co.za",
    keywords=[
        "pip",
        "datapipeline",
        "helpers",
    ],
    download_url="https://github.com/RingierIMU/sdc-global-dataPipeline-helpers/archive/v1.3.35.zip",
)



"""CUSTOM MODULE TO WRITE GOOGLE POSTMASER ANALYTICS DATA"""
# pylint: disable=too-few-public-methods, too-many-arguments
from typing import Union, List, Any, Dict, Tuple, Optional
from datetime import datetime
from sdc_dp_helpers.base_writers import BaseWriter


class GooglePostmasterWriter(BaseWriter):
    """WRITER FOR POSTMASTER DATA"""

    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False,
    ):
        self.success: List[bool] = []
        super().__init__(
            bucket=bucket,
            folder_path=folder_path,
            destination=destination,
            configs=configs,
        )

    def verify_data(
        self, payload: Union[List[Dict[Any, Any]], Dict[Any, Any], Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:
        """verifies data before writting to destination also designs the destination file path

        Args:
            payload (Dict[str, Any]): expecting a dictionary having data, date and dimension

        Raises:
            KeyError: if we do not find the exact keys we expect from the payload
            TypeError: if provided data object is not a list

        Returns:
            Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]: full destination path and
                                                                          data to be written
        """
        # confirm the payload keys are matching accurately with what is expected
        if not {"date", "brand", "data"}.issubset(set(payload.keys())):
            raise KeyError("invalid payload expecting: date, brand and data.")

        if not isinstance(payload["data"], list):
            raise TypeError("invalid data passed: expected List[Dict[Any,Any]]")

        try:
            datetime.strptime(payload["date"], "%Y%m%d")
        except ValueError as exc:
            raise ValueError(
                f"wrong date value expected format YYYYMMDD: {payload['date']}"
            ) from exc

        if not isinstance(payload["brand"], str):
            raise TypeError(f"wrong 'brand' value passed: {payload['brand']}")

        data: list = payload["data"]
        date: str = payload["date"]
        write_path: str = f"{self.folder_path}/{date}"
        return write_path, data



"""
    CUSTOM FALCON READER CLASSES
"""
# pylint: disable=too-few-public-methods,import-error,unused-import,too-many-locals,arguments-differ

from typing import Generator, List, Dict
from datetime import datetime

import requests

from sdc_dp_helpers.base_readers import BaseReader
from sdc_dp_helpers.api_utilities.date_managers import date_range_iterator
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.falcon.falcon_sdk import FalconAPICall



class FalconReader(BaseReader):
    """Falcon Reader class

    Args:
        BaseReader (_type_): API Reader Class
    """
    def __init__(self, creds_file: str, config_file=None):
        """creds_file has crecdentials config_file has api pull configs"""
        self._creds: dict = load_file(creds_file, "yml")
        self._config: dict = load_file(config_file, "yml")
        self.session: requests.Session = requests.Session()
        self.success = []
        self.service: FalconAPICall = self._get_auth()
        self.start = 0

    def _get_auth(self):
        self.service = FalconAPICall(
            session=self.session, creds=self._creds, config=self._config
        )

        return self.service

    def _query_handler(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> List[Dict]:

        data = self.service.get_data(
            start_date=start_date,
            end_date=end_date,
            network=network,
            channel_ids=channel_ids,
        )
        return data

    def run_query(self) -> Generator:
        """Calls the Query Handler"""
        endpoint_name = self._config.get("endpoint_name")
        channel_ids = self.service.get_channel_ids()
        network = self._config["networks"]

        date_iterator = date_range_iterator(
            start_date=self._config["since"],
            end_date=self._config["until"],
            interval="1_day",
            end_inclusive=False,
            time_format="%Y-%m-%dT00:00:00Z",
        )
        for start_date, end_date in date_iterator:
            payload: List = []
            payload = self._query_handler(
                start_date=start_date,
                end_date=end_date,
                network=network,
                channel_ids=channel_ids,
            )
            date = datetime.strftime(datetime.strptime(start_date, "%Y-%m-%dT%H:%M:%SZ"),"%Y-%m-%d")
            if payload:
                yield {
                    "networks": network,
                    "date": date,
                    "data": payload,
                }
                self.start = 0
                self.is_success()
            else:
                self.not_success()
                print(f"No data for endpoint {endpoint_name} for date : {date}")
